import pytest
import os

from schedules_tools import converter


DATA_DIR = 'data'
# tests
PARENT_DIRNAME = os.path.basename(os.path.dirname(os.path.realpath(__file__)))

# schedules_tools
BASE_DIR = os.path.dirname(os.path.realpath(
    os.path.join(__file__, os.pardir)))

# schedules_tools/tests
CURR_DIR = os.path.join(BASE_DIR, PARENT_DIRNAME)


class CheckTaskExistence(object):
    schedule = None

    @pytest.fixture(autouse=True)
    def setUp(self):
        super(CheckTaskExistence, self).setUp()
        input_schedule = os.path.join(CURR_DIR, DATA_DIR, 'input.tjx')
        conv = converter.ScheduleConverter()
        self.schedule = conv.import_schedule(input_schedule)

    def test_empty_check_list(self):
        check_tasks = dict()
        missing_tasks = self.schedule.check_for_taskname(check_tasks)
        assert len(missing_tasks) == 0

    def test_exact_match(self):
        # key = task name, value = match from beginning
        check_tasks = {
            'Release': False,
            'Test 2': False,
        }
        missing_tasks = self.schedule.check_for_taskname(check_tasks)
        assert len(missing_tasks) == 0

    def test_exact_not_match(self):
        # key = task name, value = match from beginning
        check_tasks = {
            'Releaseeee': False,
            'Test 2': False,
        }
        missing_tasks = self.schedule.check_for_taskname(check_tasks)
        assert len(missing_tasks) == 1
        assert 'Releaseeee' in missing_tasks

    def test_startswith_match(self):
        check_tasks = {
            'Devel': True,  # Should match 'Development'
            'T': True,  # Should match 'Test 1', 'Testing phase', ...
        }
        missing_tasks = self.schedule.check_for_taskname(check_tasks)
        assert len(missing_tasks) == 0

    def test_startswith_not_match(self):
        check_tasks = {
            'Developing something completely different': True,
            'Tradada': True,
        }
        missing_tasks = self.schedule.check_for_taskname(check_tasks)

        assert len(missing_tasks) == 2
        assert 'Developing something completely different' in missing_tasks
        assert 'Tradada' in missing_tasks

    def test_combine_exact_startswith_match(self):
        check_tasks = {
            'Release': False,
            'Test': True
        }
        missing_tasks = self.schedule.check_for_taskname(check_tasks)
        assert len(missing_tasks) == 0

    def test_combine_exact_startswith_not_matchs(self):
        check_tasks = {
            'Releaseeee': False,
            'Release': False,
            'Test': True,
            'Testnothing': True
        }
        missing_tasks = self.schedule.check_for_taskname(check_tasks)
        assert len(missing_tasks) == 2
        assert 'Releaseeee' in missing_tasks
        assert 'Testnothing' in missing_tasks



from datetime import datetime
import logging
import pytz

from lxml import etree

log = logging.getLogger(__name__)


# Handle implementation must be in format ScheduleHandler_format
# where 'format' is used as a uniq label for the format and
# 'ScheduleHandler' can be whatever.
class ScheduleHandlerBase(object):
    handle = None
    schedule = None

    # This flag indicate ability to export internal intermediate structure
    # (Schedule) into format of implementation. It's read by ScheduleConverter
    # during autodiscovery and used to provide actual help message in CLI

    # TODO: add provide_import to be complete?
    provide_export = False
    provide_changelog = False
    provide_mtime = False

    options = {}

    default_export_ext = None

    # Handlers can depend on additional python modules. We don't require from
    # users to have all of them installed if they aren't used.
    # This flag indicates that the handler can be fully utilized and there is
    # no missing dependent packages installed.
    handle_deps_satisfied = False

    def __init__(self, handle=None, schedule=None, options=dict()):
        self.schedule = schedule
        self.options = options

        # set handle last - there might be custom processing that requires options to already be set
        self.handle = handle  # 'handle' is source/target of schedule in general

    def _write_to_file(self, content, filename):
        with open(filename, 'wb') as fp:
            fp.write(content.strip().encode('UTF-8'))

    def get_handle_mtime(self):
        """ Implement only if schedule handler is able to get mtime directly
        without storage """
        raise NotImplementedError

    def handle_modified_since(self, mtime):
        """ Return boolean to be able to bypass processing """
        # Return False only when able to tell otherwise return True

        modified = True

        if isinstance(mtime, datetime):
            try:
                handle_mtime = self.get_handle_mtime()
            except NotImplementedError:
                pass

            # we're working with TZ naive dates (but in UTC)
            if handle_mtime:
                if handle_mtime.tzinfo is not None:
                    handle_mtime = handle_mtime.astimezone(pytz.utc).replace(tzinfo=None)

                if handle_mtime <= mtime:
                    modified = False

        return modified

    def get_handle_changelog(self):
        raise NotImplementedError

    # handle - file/link/smartsheet id
    def import_schedule(self):
        raise NotImplementedError

    def export_schedule(self):
        raise NotImplementedError

    def build_schedule(self):
        raise NotImplementedError

    @classmethod
    def is_valid_source(cls, handle=None):
        """Method returns True, if the specific handler is able to work with
        given handle"""
        return False

    def extract_backup(self, handle=None):
        """Prepare files which need a backup in case of external source"""
        return []


class TJXCommonMixin(object):
    default_export_ext = 'tjx'

    src_tree = None
    provide_changelog = True
    provide_mtime = True

    def _get_parsed_tree(self):
        if not self.src_tree:
            self.src_tree = etree.parse(self.handle)

        return self.src_tree

    def get_handle_changelog(self):
        # import changelog
        changelog = {}

        for log in self._get_parsed_tree().xpath('changelog/log'):
            changelog[log.get('rev')] = {
                'date': datetime.strptime(log.get('date'), '%Y/%m/%d'),
                'user': log.get('user'),
                'msg': log.text.strip(),
            }

        return changelog

    def get_handle_mtime(self):
        mtime = None

        changelog = self.get_handle_changelog()

        if changelog:
            mtime = changelog.values()[0]['date']

        return mtime



import logging

from schedules_tools.schedule_handlers import ScheduleHandlerBase


log = logging.getLogger(__name__)


class ScheduleHandler_abc(ScheduleHandlerBase):
    provide_export = True

    @staticmethod
    def is_valid_source(handle):
        return False

    # Schedule
    def export_schedule(self, output):
        with open(output, 'w+') as fd:
            fd.write(self.schedule.name)



import datetime
import json
import logging
import os

from schedules_tools.schedule_handlers import ScheduleHandlerBase
from schedules_tools.models import Schedule, Task

log = logging.getLogger(__name__)


class ScheduleHandler_json(ScheduleHandlerBase):
    provide_export = True

    handle_deps_satisfied = True

    default_export_ext = 'json'

    changelog_date_format = '%Y-%m-%d'

    @classmethod
    def is_valid_source(cls, handle=None):
        if not handle:
            handle = cls.handle
        file_ext = os.path.splitext(handle)[1]

        if file_ext != '.json':
            return False
        try:
            with open(handle) as fd:
                json.load(fd)
        except ValueError:
            return False

        return True

    @staticmethod
    def _parse_timestamp(timestamp):
        number = int(timestamp)
        return datetime.datetime.fromtimestamp(number)

    def import_schedule(self):
        with open(self.handle) as fd:
            jsonobj = json.load(fd)

        schedule = Schedule()
        schedule.dStart = self._parse_timestamp(jsonobj['start'])
        schedule.dFinish = self._parse_timestamp(jsonobj['end'])
        schedule.slug = jsonobj['slug']
        schedule.name = jsonobj['name']

        if jsonobj.get('mtime', None):
            schedule.mtime = self._parse_timestamp(jsonobj['mtime'])
        schedule.ext_attr = jsonobj.get('ext_attr', {})
        schedule.resources = jsonobj.get('resources', {})
        schedule.flags_attr_id = jsonobj.get('flags_attr_id', None)
        # schedule.id_reg is built during parsing tasks

        if 'changelog' in jsonobj:
            changelogs = {}

            for rev, record in jsonobj['changelog'].items():
                record_date = datetime.datetime.strptime(
                    record['date'], self.changelog_date_format)
                item = {
                    'user': record['user'],
                    'date':  record_date,
                    'msg': record['msg']
                }
                changelogs[rev] = item
            schedule.changelog = changelogs

        # We don't parse phases here, because we are collecting them
        # during parsing tasks itself

        for subtaskobj in jsonobj['tasks']:
            task = self.import_task_from_json(schedule, subtaskobj, None)
            schedule.tasks.append(task)
        return schedule

    def import_task_from_json(self, schedule, jsonobj, parenttask):
        task = Task(schedule)

        task.index = jsonobj['index']
        task.level = jsonobj['_level']
        task.name = jsonobj['name']
        task.slug = jsonobj['slug']
        schedule.id_reg.add(task.slug)

        task.priority = jsonobj['priority']
        task.p_complete = jsonobj['complete']
        task.milestone = False
        if jsonobj['type'] == 'Milestone':
            task.milestone = True
        task.flags = jsonobj['flags']

        if 'link' in jsonobj:
            task.link = jsonobj['link']

        if 'note' in jsonobj:
            task.note = jsonobj['note']

        task.dStart = self._parse_timestamp(jsonobj['start'])
        task.dFinish = self._parse_timestamp(jsonobj['end'])

        schedule.used_flags |= set(task.flags)

        if 'tasks' in jsonobj:
            for subtaskobj in jsonobj['tasks']:
                subtask = self.import_task_from_json(schedule, subtaskobj, task)
                task.tasks.append(subtask)

        return task

    def export_schedule(self, out_file):
        json_schedule = self.export_schedule_as_dict()
        content = json.dumps(json_schedule,
                             sort_keys=True,
                             indent=4,
                             separators=(',', ': '))

        self._write_to_file(content, out_file)

        return content

    def export_schedule_as_dict(self):
        schedule_dict = dict()
        schedule_dict['slug'] = self.schedule.slug
        schedule_dict['name'] = self.schedule.name
        schedule_dict['start'] = self.schedule.dStart.strftime('%s')
        schedule_dict['end'] = self.schedule.dFinish.strftime('%s')

        if self.schedule.mtime:
            schedule_dict['mtime'] = self.schedule.mtime.strftime('%s')

        schedule_dict['resources'] = self.schedule.resources
        schedule_dict['used_flags'] = sorted(list(self.schedule.used_flags))
        schedule_dict['ext_attr'] = self.schedule.ext_attr
        schedule_dict['flags_attr_id'] = self.schedule.flags_attr_id

        # We intentionally don't export id_reg attribute here - it's collected
        # during import

        schedule_dict['changelog'] = self.schedule.changelog
        for log in self.schedule.changelog.values():
            log['date'] = datetime.datetime.strftime(log['date'], '%Y-%m-%d')

        schedule_dict['tasks'] = []
        self.schedule.task_id_reg = set()

        for task in self.schedule.tasks:
            schedule_dict['tasks'].append(self.export_task_as_dict(task))

        return schedule_dict

    def export_task_as_dict(self, task, parent_slug=''):
        task_export = {}

        task_export['slug'] = task.slug
        task_export['index'] = task.index
        task_export['_level'] = task.level
        task_export['name'] = task.name
        task_export['priority'] = task.priority
        task_export['complete'] = task.p_complete
        task_export['type'] = task.get_type()
        task_export['flags'] = task.flags

        if task.note:
            task_export['note'] = task.note
        if task.link:
            task_export['link'] = task.link

        task_export['parentTask'] = parent_slug

        task_export['start'] = task.dStart.strftime('%s')
        task_export['end'] = task.dFinish.strftime('%s')

        if task.tasks:  # task has subtasks
            # prepare tasklist
            task_export['tasks'] = []

            for subtask in task.tasks:
                task_export['tasks'].append(self.export_task_as_dict(subtask, task.slug))

        return task_export



import datetime
import json
import pytest
import os

from schedules_tools.tests import create_test_schedule
from schedules_tools.converter import ScheduleConverter
from schedules_tools.diff import ScheduleDiff

BASE_DIR = os.path.dirname(os.path.realpath(__file__))


class TestDiff(object):
    sch1 = None
    sch2 = None

    @pytest.fixture(autouse=True)
    def setUp(self):
        self.sch1 = create_test_schedule()
        self.sch2 = create_test_schedule()

    def test_identical(self):
        diff = self.sch1.diff(self.sch2)

        assert diff == ''

    def test_sch_attrs(self):
        self.sch1.name = '_test_value 1.0.42'
        self.sch1.used_flags = set(['flag1', 'qe'])
        self.sch1.dStart = datetime.datetime(1970, 1, 1)
        self.sch1.dFinish = datetime.datetime(4970, 1, 1)
        self.sch2.changelog['today'] = 'change'

        diff = self.sch1.diff(self.sch2)

        assert diff != ''
        assert 'name' in diff
        assert 'used_flags' in diff
        assert 'dStart' in diff
        assert 'dFinish' in diff
        assert 'changelog' in diff

    def test_task_attrs(self):
        date_delta = datetime.timedelta(1)

        task = self.sch1.tasks[0].tasks[0]
        task.dStart -= date_delta
        task.dFinish += date_delta
        task.link = 'nothing'
        task.note = 'note123'

        diff = self.sch1.diff(self.sch2)

        assert 'link' in diff
        assert 'dStart' in diff
        assert 'dFinish' in diff
        assert 'note' in diff

    def test_missing_task(self):
        diff = self.sch1.diff(self.sch2)
        assert diff == ''

        self.sch2.tasks[0].tasks.pop()

        diff = self.sch1.diff(self.sch2)
        assert diff != ''

    def test_whole_days(self):
        time_delta = datetime.timedelta(minutes=1)
        self.sch2.tasks[0].tasks[1].tasks[0].dFinish -= time_delta

        diff = self.sch1.diff(self.sch2, whole_days=True)
        assert diff == ''


class TestScheduleDiff(object):
    ref_sched = 'sched_diff_reference.xml'

    SCHEDS_DIR = os.path.join(BASE_DIR, 'schedule_files/schedule_diff')
    OUTPUT_FILES_DIR = os.path.join(BASE_DIR, 'fixtures/schedule_diff')

    diff_test_scenarios = (
        ('sched_diff_dates_changed', False, False),
        ('sched_diff_subtask_added', False, False),
        ('sched_diff_subtask_removed', False, False),
        ('sched_diff_subtask_removed', True, False),
        ('sched_diff_subtask_tree_added', False, False),
        ('sched_diff_subtask_tree_removed', False, False),
        ('sched_diff_task_appended', False, False),
        ('sched_diff_task_appended', True, False),
        ('sched_diff_tasks_renamed', False, False),
        ('sched_diff_time_changed', False, False),
        ('sched_diff_time_changed', True, False),
        ('sched_diff_reference', True, True),
        ('sched_diff_flags_changed', True, True),
        ('sched_diff_flags_changed', True, False),
    )

    def import_schedule(self, filename):
        path = os.path.join(self.SCHEDS_DIR, filename)
        conv = ScheduleConverter()
        return conv.import_schedule(path)

    @pytest.mark.parametrize(
        'schedule_name,trim_time,compare_flags',
        diff_test_scenarios
    )
    @pytest.mark.parametrize(
        'output_format',
        ['json', 'txt']
    )
    def test_diff(self, schedule_name, trim_time, compare_flags, output_format):

        left = self.import_schedule(self.ref_sched)
        right = self.import_schedule('%s.xml' % schedule_name)

        if compare_flags:
            extra_compare_attributes = ['flags']
        else:
            extra_compare_attributes = None

        diff = ScheduleDiff(
            left,
            right,
            trim_time=trim_time,
            extra_compare_attributes=extra_compare_attributes
        )

        if output_format == 'json':
            diff_output = diff.dump_json(indent=4)
        elif output_format == 'txt':
            diff_output = str(diff)
        else:
            raise ValueError

        diff_reference_filename = '%s%s%s.%s' % (
            schedule_name,
            '_trim_time' if trim_time else '',
            '_compare_flags' if compare_flags else '',
            output_format
        )

        regenerate = os.environ.get('REGENERATE', 'false').lower() == 'true'
        if regenerate:
            file_mode = 'w'
        else:
            file_mode = 'r'

        with open(
            os.path.join(self.OUTPUT_FILES_DIR, diff_reference_filename),
            file_mode
        ) as f:
            if regenerate:
                f.write(diff_output)
            else:
                expected_diff_output = f.read()

                if output_format == 'json':
                    # Need to load them again:
                    # 1) Dates are saved as strings in expected
                    # 2) Dict equality doesn't depend on keys order
                    expected_diff_output = json.loads(expected_diff_output)
                    diff_output = json.loads(diff_output)

                assert diff_output == expected_diff_output


class TestDiffCLI(object):
    pass



#!/usr/bin/env python3

"""
Simple wrapper to get diff of two schedules

It's able to show different attributes (by 'attrs' kwarg)
and indicate missing phases

Follows 'diff' exit codes:
    0 - same
    1 - different
    2 - other trouble
"""

import argparse
import logging
import sys

from schedules_tools import discovery
from schedules_tools.diff import ScheduleDiff

import schedules_tools.converter as conv


def setup_logging(level):
    log_format = '%(name)-10s %(levelname)7s: %(message)s'
    sh = logging.StreamHandler(sys.stdout)
    sh.setLevel(level)

    formatter = logging.Formatter(log_format)
    sh.setFormatter(formatter)

    # setup root logger
    inst = logging.getLogger('')
    inst.setLevel(level)
    inst.addHandler(sh)


def main():
    setup_logging(logging.DEBUG)
    parser = argparse.ArgumentParser(
        description='Tool to show differences between two schedules.')

    parser.add_argument('--simple-diff',
                        help='Simple comparison between two schedules.',
                        action='store_true',
                        default=False)

    parser.add_argument(
        '--handlers-path',
        help='Add python-dot-notation path to discover handlers (needs to '
             'be python module), can be called several times '
             '(conflicting names will be overriden - the last '
             'implementation will be used)',
        action='append',
        default=[])
    parser.add_argument('--whole-days',
                        help='Compare just date part of timestamp (will '
                             'ignore differences in time)',
                        action='store_true',
                        default=False)
    parser.add_argument('left')
    parser.add_argument('right')
    args = parser.parse_args()

    for path in args.handlers_path:
        discovery.search_paths.append(path)

    left = conv.ScheduleConverter()
    left.import_schedule(args.left)

    right = conv.ScheduleConverter()
    right.import_schedule(args.right)

    if args.simple_diff:
        diff_res = left.schedule.diff(right.schedule, whole_days=args.whole_days)
    else:
        diff_res = ScheduleDiff(left.schedule, right.schedule)

    if diff_res:
        print(diff_res)
        sys.exit(1)


if __name__ == '__main__':
    main()



import grp
import logging
import os
import re
import shutil
import subprocess
import tempfile

import datetime as datetime_mod

from distutils.dir_util import remove_tree, copy_tree

from schedules_tools import SchedulesToolsException
from schedules_tools.storage_handlers import StorageBase

log = logging.getLogger(__name__)


class CvsCommandException(SchedulesToolsException):
    pass


class NotImplementedFeature(SchedulesToolsException):
    pass


class StorageHandler_cvs(StorageBase):
    provide_mtime = True
    provide_changelog = True

    checkout_dir = None  # path to shared local working copy of repository
    repo_root = None  # :pserver:$USER@cvs.myserver.com:/cvs/reporoot
    repo_name = None  # repo

    tmp_root = None  # local tmp handle dir

    refresh_validity = 5  # seconds
    _last_refresh_local = None
    block_refresh = False  # allows to skip refresh

    local_handle = None

    exclusive_access_option = 'exclusive_access'

    def __init__(self, handle=None, options=dict(), **kwargs):
        self.checkout_dir = options.get('cvs_checkout_path')
        self.checkout_dir_perm = options.get('cvs_checkout_dir_permission',
                                             None)
        self.repo_name = options.get('cvs_repo_name')
        self.repo_root = options.get('cvs_root')
        self.block_refresh = options.get('cvs_block_refresh', False)

        # local checkout access
        self.group_owner = options.get('cvs_group_owner', None)  # name, not id
        self.group_writable = options.get('cvs_group_writable', True)

        super(StorageHandler_cvs, self).__init__(handle, options, **kwargs)

    @property
    def redis_key(self):
        return '_'.join([super(StorageHandler_cvs, self).redis_key,
                         self.repo_root,
                         self.repo_name,
                         self.checkout_dir])

    def _cvs_command(self, cmd,
                     stdout=subprocess.PIPE,
                     stderr=subprocess.PIPE,
                     cwd=None,
                     exclusive=False):
        if not cwd:
            cwd = self.checkout_dir
        # -q, make cvs more quiet
        # -z9, maximum compression
        # -d, set CVSROOT
        cmd_str = 'cvs -q -z9 -d {} {}'.format(self.repo_root, cmd)

        if exclusive:
            self.acquire_shared_lock(msg='command "{}"'.format(cmd_str))

        log.debug('CVS command: {} (cwd={})'.format(cmd_str, cwd))
        p = subprocess.Popen(cmd_str.split(),
                             stdout=stdout,
                             stderr=stderr,
                             cwd=cwd)
        stdout, stderr = p.communicate()

        if exclusive:
            self.release_shared_lock()

        if p.returncode != 0:
            msg = str({'stdout': stdout, 'stderr': stderr, 'cmd': cmd_str})
            raise CvsCommandException(msg, source=self)

        return stdout, stderr

    @staticmethod
    def _wipe_out_dir(target_dir):
        if os.path.exists(target_dir):
            log.debug('Wipping out {}'.format(target_dir))
            remove_tree(target_dir)

    def _is_valid_cvs_dir(self, path):
        """
        Verifies that given (local) path comes from CVS local working copy
         - contains CVS directory
         - <path>/CVS/Root ends like: '@cvs.myserver.com:/cvs/reporoot'
         - <path>/CVS/Repository 'repo/' - path to tracked directory

        Args:
            path: local path of versioned directory (/tmp/cvsXZY/repo/schedule)

        Returns:
            True, if the path come from the already checked out repo
        """
        def get_subpath(*subpath):
            return os.path.join(path, *subpath)

        def cvs_root_parser(string):
            # :gserver:$USER@cvs.myserver.com:/cvs/reporoot or local /cvs/root
            chunks = string.split(':')
            if len(chunks) == 1:
                server_url = ''
                server_path = chunks[0]
            else:
                server_url = chunks[2].split('@')[1]
                server_path = chunks[3]
            return server_url, server_path

        if not os.path.exists(path):
            return False

        # CVS has to be a directory
        if not os.path.isdir(get_subpath('CVS')):
            return False

        # verify CVS module name (repo_name)
        with open(get_subpath('CVS', 'Repository')) as fd:
            line = fd.readline().strip()
            if not line.startswith(self.repo_name):
                return False

        # verify CVS Root
        with open(get_subpath('CVS', 'Root')) as fd:
            line = fd.readline().strip()
            if cvs_root_parser(line) != cvs_root_parser(self.repo_root):
                return False

        return True

    def get_local_shared_path(self, path):
        """
        Returns path to file that belongs local working copy of repository

        Args:
            path: relative path in repo (repo/schedule)

        Returns:
            Real local path to given file within repo (/tmp/tmpXYZ/repo/schedule)
        """
        ret = os.path.join(self.checkout_dir, self.repo_name, path)
        return os.path.realpath(ret)

    def refresh_local(self, force=False):
        """
        Check if local copy exists (not just path but also CVS content) and
        decide if checkout is needed or not
        If empty:
            do a cvs checkout to mkdtmp dir and then copy to local path
        If exists:
            do a cvs update - handle conflicts,...
        """
        cvs_content_root_dir = os.path.join(self.checkout_dir, self.repo_name)

        if not self._is_valid_cvs_dir(cvs_content_root_dir):
            self._cvs_checkout()

        elif not self.block_refresh:
            # use redis key if possible to share value across workers
            datetime_format = '%Y-%m-%d %H:%M:%S'
            last_refresh_local_time_key = self.redis_key + '_last_refresh_local'
            last_refresh_expired = True  # init

            last_refresh_time = self._last_refresh_local

            if self.redis:
                r_value = self.redis.get(last_refresh_local_time_key)
                if r_value:
                    last_refresh_time = str(r_value, 'utf-8')

            if last_refresh_time:
                last_refresh_time = datetime_mod.datetime.strptime(last_refresh_time,
                                                                   datetime_format)
                last_refresh_expired = (
                    last_refresh_time <
                    datetime_mod.datetime.now() - datetime_mod.timedelta(
                                                                seconds=self.refresh_validity)
                )

            if force or last_refresh_expired:
                self._update_shared_repo()

                last_refresh_time = datetime_mod.datetime.now().strftime(datetime_format)
                self._last_refresh_local = last_refresh_time

                if self.redis:
                    self.redis.set(last_refresh_local_time_key, last_refresh_time)

    def _copy_subtree_to_tmp(self, processed_path):
        """
        Create an independent copy of schedule (from main-cvs-checkout),
        located in /tmp and

        Args:
            processed_path: relative subtree path of shared checkout dir

        Returns:
            Path to processed_path copied directory  in /tmp
        """
        src = os.path.join(self.checkout_dir, self.repo_name, processed_path)

        dst_tmp_dir = tempfile.mkdtemp(prefix='sch_')
        dst = os.path.join(dst_tmp_dir, processed_path)

        # lock cvs so nothing changes during copy
        self.acquire_shared_lock(msg='copying subtree from shared cvs copy')
        copy_tree(src, dst)
        self.release_shared_lock()

        return dst_tmp_dir

    def _process_path(self):
        """
        Get handle and figure out, which directory subtree needs to be check
        out
        Example: schedule/schedule-1-0.tjp -> schedule

        Returns:
            Path to schedule, that handle is part of ('schedule')
        """
        return os.path.dirname(self.handle)

    def get_local_handle(self, enforce=False, revision=None, datetime=None):
        """
        Facilitate update/clone of shared working copy of repository,

        Args:
            revision:
            datetime:

        Returns:
            Tuple (path_to_handle, path_to_tmp_dir)

        """
        if not self.local_handle or enforce:
            # Checkout exact revision/datetime if defined
            if revision or datetime:
                raise NotImplementedFeature(
                    'Checking out specific revision is not currently '
                    'implemented.')

            self.refresh_local()  # refresh shared local tree
            subtree_rel_path = self._process_path()
            self.tmp_root = self._copy_subtree_to_tmp(subtree_rel_path)

            self.local_handle = os.path.join(self.tmp_root, self.handle)
            log.debug('Creating local handle {}'.format(self.local_handle))

        return self.local_handle

    def clean_local_handle(self):
        if self.local_handle:
            log.debug('Cleaning local handle {}'.format(self.local_handle))

            remove_tree(self.tmp_root)
            self.local_handle = None

    def _cvs_checkout(self, force=False):
        """
        Clones the repo into:
            1) cvs_checkout_path optional param
            2) made temp directory

        Set self.checkout_dir to new, proper path accordingly.
        """
        verify_existing_dif = True

        if self.group_writable:
            orig_umask = os.umask(0o002)  # group writable

        if not self.checkout_dir:
            verify_existing_dif = False
            self.checkout_dir = tempfile.mkdtemp(prefix='sch_repo_')

        log.debug('Using {} as checkout dir'.format(self.checkout_dir))

        if force:
            verify_existing_dif = False

        local_repo_path = self.get_local_shared_path(self.repo_name)
        if verify_existing_dif and self._is_valid_cvs_dir(local_repo_path):
            return self.checkout_dir

        temp_checkout_dir = tempfile.mkdtemp(prefix='sch_tmp_repo_')

        if self.group_owner:
            os.chown(temp_checkout_dir, -1, grp.getgrnam(self.group_owner).gr_gid)

        if self.checkout_dir_perm:
            os.chmod(temp_checkout_dir, self.checkout_dir_perm)

        cmd = 'co {}'.format(self.repo_name)

        self._cvs_command(cmd, cwd=temp_checkout_dir, exclusive=True)

        # Finally move whole content into required destination. Due behavior of
        # 'move', the destination don't have to exists. That's why it can't be
        # done as atomic operation.
        # 'atomic' part start
        self._wipe_out_dir(self.checkout_dir)
        shutil.move(temp_checkout_dir, self.checkout_dir)
        # 'atomic' part end

        if self.checkout_dir_perm:
            os.chmod(self.checkout_dir, self.checkout_dir_perm)

        if orig_umask:
            os.umask(orig_umask)  # group writable

        return self.checkout_dir

    def _cvs_update(self, filename='', revision=None, datetime_rev=None):
        """
        Run 'cvs update' with given filename or if it's not specified whole
        local working copy of repository and specification of revision
        to get, if it's specified.

        Args:
            filename: relative path (repo/schedule/schedule-1-0)
            revision: Optional, pick specific revision by revision number
            datetime_rev: Optional, pick specific revision by date

        Returns:
            stdout/err of 'cvs' command
        """
        cmd_revision = ''
        cvs_params = ''
        # -d = retrieve also new directories
        # -P = prune empty directories
        cmd_params = '-dP'

        if revision and datetime_rev:
            raise CvsCommandException('Revision specification by number and '
                                      'date are disjunctive. Pick just one.',
                                      source=self)

        if datetime_rev:
            datetime_str = datetime_rev.strftime('%Y-%m-%d %H:%M')
            cmd_revision = '-D "{}"'.format(datetime_str)
        if revision:
            cmd_revision = '-r "{}"'.format(revision)

        if self.group_writable:
            orig_umask = os.umask(0o002)  # group writable

        cmd = '{cvs_params} update {cmd_params} {cmd_revision} {filename}'.format(
            cvs_params=cvs_params,
            cmd_params=cmd_params,
            cmd_revision=cmd_revision, filename=filename).strip()
        stdout, stderr = self._cvs_command(cmd, exclusive=True)

        if orig_umask:
            os.umask(orig_umask)  # group writable

        return stdout, stderr

    def _update_shared_repo(self):
        """
        Wraps 'cvs update' and in case of problems, fix the shared working copy
        of the repo (remove the dir, checkout it again), or as a last chance,
        do complete checkout of the repository from scratch.
        """
        try:
            stdout, _ = self._cvs_update()

            # Are there any directories with troubles? Remove them (afterwards).
            to_cleanup = self._parse_cvs_update_output(stdout)

            # There are no problems - nothing to do, leaving.
            if not to_cleanup:
                return

            for path in to_cleanup:
                self._clean_checkout_directory(path)

            # This step is just for sure - shouldn't occur usually
            stdout, _ = self._cvs_update()
            to_cleanup = self._parse_cvs_update_output(stdout)
            if to_cleanup:
                # Something went wrong by update, do complete fresh checkout
                self._cvs_checkout(force=True)
        except CvsCommandException as e:
            log.exception(e)
            self._cvs_checkout(force=True)

    @staticmethod
    def _parse_cvs_update_output(cvsoutput):
        """
        Parse line by line from given argument and returns list of paths, that's
        somehow corrupted and have to be updated/checkouted from scratch.

        Args:
            cvsoutput: string with (usually) multiple lines, stdout of
                       'cvs update' command

        Returns:
            list of paths, that's corrupted

        """
        re_flags = re.compile(r'^(\S)\s+(\S+)')
        to_cleanup = set()

        for line in cvsoutput.splitlines():
            match = re.findall(re_flags, str(line, 'utf-8'))
            if not match:
                continue
            flag, filename = match[0]
            path = os.path.dirname(filename)
            # Conflict, Added, Modified locally, ? not versioned yet
            if flag in ('C', 'A', 'M', '?'):
                if path != '':
                    to_cleanup.add(path)
            # Updated, Patched
            elif flag in ('U', 'P'):
                # it's valid state - there exists remote changes,
                # local working copy doesn't have any external changes
                pass
            else:
                log.debug('Unknown CVS status flag: ' + flag)
                # Don't know what it means, so clean it to be sure
                if path != '':
                    to_cleanup.add(path)
        return to_cleanup

    def _clean_checkout_directory(self, directory):
        log.debug('Wipe out checkout directory {}'.format(directory))
        rm_dir = os.path.join(self.checkout_dir, directory)
        if os.path.exists(rm_dir):
            # lock cvs so nothing changes during cleanup
            self.acquire_shared_lock(msg='cleaning shared cvs copy')
            remove_tree(rm_dir)
            self.release_shared_lock()

    def get_handle_mtime(self):
        self.refresh_local()

        local_handle = self.get_local_shared_path(self.handle)
        mtime_timestamp = os.path.getmtime(local_handle)
        return datetime_mod.datetime.fromtimestamp(mtime_timestamp).replace(microsecond=0)

    def get_handle_changelog(self):
        self.refresh_local()

        changelog = {}
        cmd = 'log {}'.format(os.path.join(self.repo_name, self.handle))
        stdout, _ = self._cvs_command(cmd)

        STATE_HEAD = 'head'
        STATE_REVISION = 'revision'
        STATE_DATE_AUTHOR = 'dateauthor'
        STATE_COMMENT = 'comment'

        state = STATE_HEAD

        revision = None
        author = None
        date = None
        comment = []
        re_date_author = re.compile(r'date:\s*([^;]+);\s+author:\s*([^;]+);')
        re_head = re.compile(r'head:\s+(.+)')
        re_revision = re.compile(r'revision\s+(.+)')
        re_branches = re.compile(r'branches:\s+(.+);')
        log_separator = '----------------------------'
        log_end_separator = \
            '============================================================================='

        for line in stdout.splitlines():
            line = str(line, 'utf-8')
            if state == STATE_HEAD:
                matches = re_head.findall(line)
                if not matches:
                    continue

                state = STATE_REVISION
                continue
            elif state == STATE_REVISION:
                # new record, clean all previous values
                matches = re_revision.findall(line)
                if not matches:
                    continue
                revision = matches[0]
                author = None
                date = None
                comment = []

                state = STATE_DATE_AUTHOR
                continue
            elif state == STATE_DATE_AUTHOR:
                matches = re_date_author.findall(line)
                date = datetime_mod.datetime.strptime(
                    matches[0][0], '%Y/%m/%d %H:%M:%S')
                author = matches[0][1]
                comment = []

                state = STATE_COMMENT
                continue
            elif state == STATE_COMMENT:
                br = re_branches.match(line)
                if not comment and br:
                    continue
                if line == log_separator or line == log_end_separator:
                    # store whole log
                    comment = '\n'.join(comment)
                    record = {
                        'user': author,
                        'date': date,
                        'msg': comment
                     }
                    changelog[revision] = record
                    state = STATE_REVISION
                    continue
                comment.append(line)
                continue

        return changelog



from schedules_tools.schedule_handlers import ScheduleHandlerBase
import logging
from lxml.html import etree

log = logging.getLogger(__name__)


class ScheduleHandler_confluencehtml(ScheduleHandlerBase):
    provide_export = True

    handle_deps_satisfied = True

    default_export_ext = 'html'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.options.get('date_format', False):
            self.options['date_format'] = '%a %Y-%m-%d'

    @classmethod
    def is_valid_source(cls, handle=None):
        return False

    def _export_task(self, e_parent, task, top=True):

        if top:  # top level - make p
            e_p = etree.SubElement(e_parent, 'p')
            e_strong = etree.SubElement(e_p, 'strong')
            e_strong.text = task.name
        else:
            e_li = etree.SubElement(e_parent, 'li')
            e_li.text = f'{task.name}'
            e_li.attrib['role'] = 'checkbox'

            if hasattr(task, 'user'):
                e_user = etree.SubElement(e_li, 'em')
                e_user.text = f' [{task.user}]'

        if len(task.tasks):
            e_ul = etree.SubElement(e_parent, 'ul')
            e_ul.attrib['class'] = 'inline-task-list'

            for task in task.tasks:
                self._export_task(e_ul, task, top=False)

        if top:
            etree.SubElement(e_parent, 'br')

    # Schedule
    def export_schedule(self, out_file=None):
        e_html = etree.Element('html')
        e_head = etree.SubElement(e_html, 'head')

        etree.SubElement(e_head, 'meta', charset="utf-8")

        if self.options.get('html_title', False):
            title = self.options['html_title']
        else:
            title = self.schedule.name

        title_date_fmt = '%b %-d'
        start_date = self.schedule.dStart.strftime(title_date_fmt)
        finish_date = self.schedule.dFinish.strftime(title_date_fmt)
        title_text = f'{title} ({start_date} - {finish_date})'

        e_title = etree.SubElement(e_head, 'title')
        e_title.text = title_text

        e_body = etree.SubElement(e_html, 'body')

        e_h = etree.SubElement(e_body, 'h1')
        e_h.text = title_text

        etree.SubElement(e_body, 'br')

        for task in self.schedule.tasks:
            self._export_task(e_body, task)

        etree_return = etree.ElementTree(e_html)
        if out_file:
            etree_return.write(out_file, pretty_print=True, encoding="utf-8",
                               xml_declaration=False)

        return str(etree_return)



from schedules_tools import models
import datetime


def create_test_schedule():
    sch = models.Schedule()
    sch.dStart = datetime.datetime(2016, 12, 1, 8, 0)
    sch.dFinish = datetime.datetime(2016, 12, 22, 8, 0)
    sch.name = 'Schedules tools'
    sch.used_flags = set(['qe', 'pm', 'dev'])
    sch.slug = 'scht'

    task1 = models.Task(sch)
    task1.name = 'Schedules tools 1.0.0'
    task1.slug = 'scht'
    task1.dStart = datetime.datetime(2016, 12, 1, 8, 0)
    task1.dFinish = datetime.datetime(2016, 12, 20, 8, 0)
    task1.note = 'top task'
    task1.priority = 13
    task1.p_complete = 0
    sch.tasks.append(task1)

    task2 = models.Task(sch, level=task1.level + 1)
    task2.name = 'Planning'
    task2.slug = 'scht.plan'
    task2.dStart = datetime.datetime(2016, 12, 2, 8, 0)
    task2.dFinish = datetime.datetime(2016, 12, 4, 8, 0)
    task2.note = 'plan task'
    task2.priority = 34
    task2.p_complete = 10
    task2.flags = ['pm']
    task2.link = 'https://www.redhat.com'
    task1.tasks.append(task2)

    task3 = models.Task(sch, level=task1.level + 1)
    task3.name = 'Development'
    task3.slug = 'scht.dev'
    task3.dStart = datetime.datetime(2016, 12, 5, 8, 0)
    task3.dFinish = datetime.datetime(2016, 12, 15, 8, 0)
    task3.note = 'devel task'
    task3.priority = 100
    task3.p_complete = 90
    task3.flags = ['dev']
    task3.milestone = 'dev-milestone'
    task1.tasks.append(task3)

    task4 = models.Task(sch, level=task3.level + 1)
    task4.name = 'Handlers'
    task4.slug = 'scht.dev.handlers'
    task4.dStart = datetime.datetime(2016, 12, 10, 8, 0)
    task4.dFinish = datetime.datetime(2016, 12, 12, 8, 0)
    task4.priority = 100
    task4.p_complete = 123
    task4.flags = ['dev', 'qe']
    task3.tasks.append(task4)

    return sch
